"""Basic stuff for implementing pull plugins."""

import multiprocessing as proc
import time
from abc import abstractmethod
from datetime import datetime

from schedule import Scheduler

from .. import Plugin
from ...utils import (auto_str_ignore, parse_duration_literal, StopCycleError,
                      interruptible_sleep, try_parse_bool)


@auto_str_ignore(['stopped', '_stopped', 'on_payload'])
class PullBase(Plugin):
    """
    Base class for pull plugins.
    The plugin has to implements `pull` to do the actual data retrieval / task.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._stopped = proc.Event()

    @property
    def stopped(self):
        """Returns True if the pull is considered stopped; otherwise False."""
        return self._stopped.is_set()

    @abstractmethod
    def pull(self):
        """
        Performs the actual data retrieval / task.
        Returns:
            None.
        """
        raise NotImplementedError()  # pragma: no cover

    def on_payload(self, payload):
        """Callback for execution engine."""

    def notify(self, payload):
        """Call in subclass to emit some payload to the execution engine."""
        self.on_payload(self, payload)  # pylint: disable=too-many-function-args

    def stop(self):
        """Stops the plugin."""
        self._stopped.set()

    def _sleep(self, sleep_time=10):
        """Call in subclass to perform some sleeping."""
        def callback():
            if self.stopped:
                raise StopCycleError()
        interruptible_sleep(sleep_time, callback, interval=0.5)


class PollingError(Exception):
    """Base class for errors that occur during polling."""


class StopPollingError(Exception):
    """Raise this exception in child classes if you want to stop the polling scheduler."""


@auto_str_ignore(['_scheduler'])
class Polling(PullBase):
    """
    Base class for polling plugins.

    You may specify duration literals such as 60 (60 secs), 1m, 1h (...) to realize a periodic
    polling or cron expressions (*/1 * * * * > every min) to realize cron like behaviour.
    """
    __prefix__ = 'poll'

    def __init__(self, interval=60, instant_run=False, **kwargs):
        super().__init__(**kwargs)
        try:
            # Literals such as 60s, 1m, 1h, ...
            self._interval = parse_duration_literal(interval)
            self.is_cron = False
        except TypeError:
            # ... or a cron-like expression is valid
            from cronex import CronExpression
            self._interval = CronExpression(interval)
            self.is_cron = True

        self._scheduler = None
        self._instant_run = try_parse_bool(instant_run, False)

    def pull(self):
        self._scheduler = Scheduler()
        self._configure_scheduler(self._scheduler, self._run_schedule)

        if self._instant_run:
            self._scheduler.run_all()

        while not self.stopped:
            self._scheduler.run_pending()
            time.sleep(0.5)

    def _run_schedule(self):
        try:
            if self.is_cron:
                dtime = datetime.now()
                if not self._interval.check_trigger((dtime.year, dtime.month,
                                                     dtime.day, dtime.hour, dtime.minute)):
                    return  # It is not the time for the cron to trigger
            payload = self.poll()
            if payload is not None:
                self.notify(payload)
        except StopPollingError:
            self.stop()
        except Exception:  # pragma: no cover, pylint: disable=broad-except
            import traceback
            self.logger.error("[%s]\n%s", self.name, traceback.format_exc())

    def _configure_scheduler(self, scheduler, callback):
        """
        Configures the scheduler. You have to differ between "normal" intervals and
        cron like expressions by checking `self.is_cron`.

        Override in subclasses to fir the behaviour to your needs.
        Args:
            scheduler (schedule.Scheduler): The actual scheduler.
            callback (callable): The callback to call when the time is right.

        Returns:
            None

        """
        if self.is_cron:
            # Scheduler always executes at the exact minute to check for cron triggering
            scheduler.every().minute.at(":00").do(callback)
        else:
            # Scheduler executes every interval seconds to execute the poll
            scheduler.every(self._interval).seconds.do(callback)

    @abstractmethod
    def poll(self):
        """
        Implement in plugin components to do the actual polling.

        Returns: Returns the data for the downstream pipeline.

        """
        raise NotImplementedError()  # pragma: no cover



"""Simple pull plugins"""

import sys
import time
from datetime import datetime

from . import PullBase, Polling
from ...utils import make_list, auto_str_ignore
from ...validator import Validator


class Count(PullBase):
    """
    Emits every `wait` seconds a counting value which runs from `from_cnt` to `to_cnt`.
    If `to_cnt` is None will to count to infinity.

    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/pull/simple.Count/index.md

    """

    def __init__(self, from_cnt=0, to_cnt=None, wait=5, **kwargs):
        super().__init__(**kwargs)
        self.from_cnt = int(from_cnt)
        self.to_cnt = to_cnt and int(to_cnt)
        self.wait = float(wait)

    def pull(self):
        for i in range(self.from_cnt, self.to_cnt or sys.maxsize):
            self._sleep(self.wait)
            self.notify(i)
            if self.stopped:
                break


@auto_str_ignore(['jobs'])
class Cron(Polling):
    """
    Cron-like triggering of dependent pushes.

    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/pull/simple.Cron/index.md
    """

    def __init__(self, expressions, **kwargs):
        super().__init__(interval=60, instant_run=False, **kwargs)
        self.expressions = make_list(expressions)

        from cronex import CronExpression
        self.jobs = [CronExpression(expression) for expression in self.expressions]

    def _configure_scheduler(self, scheduler, callback):
        scheduler.every().minute.at(":00").do(callback)

    def poll(self):
        for job in self.jobs:
            dtime = datetime.now()
            if job.check_trigger((dtime.year, dtime.month, dtime.day, dtime.hour, dtime.minute)):
                self.notify({'data': job.comment})


class CustomPolling(Polling):
    """
    Calls the specified callable every `interval`. The result of the callable is simply returned.
    This plugin is basically for _internal_ use only.
    """
    def __init__(self, scheduled_callable, **kwargs):
        super().__init__(**kwargs)
        self.scheduled_callable = scheduled_callable
        Validator.is_function(scheduled_callable=self.scheduled_callable)

    def poll(self):
        return self.scheduled_callable()


class Infinite(PullBase):
    """Just for demonstration purposes. DO NOT USE!"""

    def __init__(self, **kwargs):  # pragma: no cover, pylint: disable=useless-super-delegation
        super().__init__(**kwargs)

    def pull(self):  # pragma: no cover
        while True:
            time.sleep(0.5)


class Repeat(PullBase):
    """
    Emits every `wait` seconds the same `repeat`.

    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/pull/simple.Repeat/index.md

    """

    def __init__(self, repeat, wait=5, **kwargs):
        super().__init__(**kwargs)
        self.repeat = repeat
        self.wait = float(wait)

    def pull(self):
        while not self.stopped:
            self._sleep(self.wait)
            self.notify(self.repeat)



"""Camera related pull plugins"""

import os

from .fs import FileSystemWatcher
from ...utils import parse_duration_literal, Debounce, auto_str_ignore
from ...validator import Validator


@auto_str_ignore(['_debouncer'])
class MotionEyeWatcher(FileSystemWatcher):
    """
    Watches a MotionEye directory (where the images and the movies get persisted from motioneye)
    to trigger some useful events based on new / modified files.
    The motion event only works, when the camera is configured to persist images / movies
    only when motion is triggered and not 24/7.

    See Also:
        https://github.com/HazardDede/pnp/blob/feature/motioneye/docs/plugins/pull/camera.MotionEyeWatcher/index.md

    """

    CONF_EVENT_TYPE_IMAGE = 'image'
    CONF_EVENT_TYPE_MOVIE = 'movie'
    CONF_EVENT_TYPE_MOTION = 'motion'
    CONF_MOTION_ON = 'on'
    CONF_MOTION_OFF = 'off'

    def __init__(self, path, image_ext='jpg', movie_ext='mp4', motion_cool_down='30s', **kwargs):
        if not image_ext and not movie_ext:
            raise TypeError("You have to specify either `image_file_ext`, `movie_file_ext` or both")

        Validator.is_instance(str, allow_none=True, image_ext=image_ext)
        self.image_ext = image_ext
        if self.image_ext:
            self.image_ext = self.image_ext if self.image_ext.startswith('.') else '.' + self.image_ext
        Validator.is_instance(str, allow_none=True, movie_ext=movie_ext)
        self.movie_ext = movie_ext
        if self.movie_ext:
            self.movie_ext = self.movie_ext if self.movie_ext.startswith('.') else '.' + self.movie_ext

        self._events = []
        self._patterns = []
        if self.image_ext:
            self._patterns.append('*{}'.format(self.image_ext))
            self._events.append(self.EVENT_TYPE_CREATED)
        if self.movie_ext:
            self._patterns.append('*{}'.format(self.movie_ext))
            self._events.append(self.EVENT_TYPE_MODIFIED)

        super().__init__(
            path=path,
            recursive=True,
            patterns=self._patterns,
            events=self._events,
            load_file=False,
            ignore_directories=True,
            **kwargs
        )

        self.motion_cool_down = parse_duration_literal(motion_cool_down)
        self._debouncer = None

    def _motion_off(self, notify_impl):
        self.logger.debug("Send motion 'off' (initial)")
        notify_impl(dict(event=self.CONF_EVENT_TYPE_MOTION, state=self.CONF_MOTION_OFF))
        self._debouncer = None

    def _motion_on(self, notify_impl):
        if self._debouncer is None:
            self.logger.debug("Send motion 'on' (initial)")
            notify_impl(dict(event=self.CONF_EVENT_TYPE_MOTION, state=self.CONF_MOTION_ON))
            self._debouncer = Debounce(self._motion_off, self.motion_cool_down)
        else:
            self.logger.debug("Send motion 'on' (debounced)")
        self._debouncer(notify_impl=notify_impl)

    def notify(self, payload):
        operator = payload.get('operation')
        source = payload.get('source')
        _, ext = os.path.splitext(source)

        if operator == self.EVENT_TYPE_CREATED and ext == self.image_ext:  # This one's an image
            self.logger.debug("Got image '%s' (created)", source)
            self._motion_on(super().notify)
            super().notify(payload=dict(event=self.CONF_EVENT_TYPE_IMAGE, source=source))
        elif operator == self.EVENT_TYPE_CREATED and ext == self.movie_ext:  # Creation of movie -> trigger motion
            self.logger.debug("Got movie '%s' (created)", source)
            self._motion_on(super().notify)
        elif operator == self.EVENT_TYPE_MODIFIED and ext == self.movie_ext:  # This one's a movie
            self.logger.debug("Got movie '%s' (modified)", source)
            self._motion_on(super().notify)
            super().notify(payload=dict(event=self.CONF_EVENT_TYPE_MOVIE, source=source))
        else:
            self.logger.debug("Got event '%s' with source '%s', but not supported. Ignored...", operator, source)

    def stop(self):
        if self._debouncer:
            self._debouncer.execute_now()
        super().stop()



import mimetypes
import os
from email import encoders
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

from ..validator import Validator


class Mail:
    @staticmethod
    def _content_type(file_path):
        content_type, encoding = mimetypes.guess_type(file_path)
        if content_type is None or encoding is not None:
            content_type = 'application/octet-stream'
        main_type, sub_type = content_type.split('/', 1)
        return main_type, sub_type

    @staticmethod
    def _file_contents(file_path, file_type):
        with open(file_path, 'r' if file_type == 'text' else 'rb') as fp:
            return fp.read()

    @staticmethod
    def _make_attachment(file_path, contents, main_type, sub_type):
        if main_type == 'text':
            from email.mime.text import MIMEText
            msg = MIMEText(contents, _subtype=sub_type)
        elif main_type == 'image':
            from email.mime.image import MIMEImage
            msg = MIMEImage(contents, _subtype=sub_type)
        elif main_type == 'audio':
            from email.mime.audio import MIMEAudio
            msg = MIMEAudio(contents, _subtype=sub_type)
        else:
            from email.mime.base import MIMEBase
            msg = MIMEBase(main_type, sub_type)
            msg.set_payload(contents)

        encoders.encode_base64(msg)
        filename = os.path.basename(file_path)
        msg.add_header('Content-Disposition', 'attachment', filename=filename)
        return msg

    @classmethod
    def create_with_attachment(cls, sender, to, subject, file_path, message_text=None):
        """Create a message for an email.

        Example:

            >>> import tempfile
            >>> with tempfile.NamedTemporaryFile() as attach:
            ...     res = Mail.create_with_attachment('sender', 'a@a.com', 'subject', attach.name, 'Message text')
            >>> res['to'], res['from'], res['subject']
            ('a@a.com', 'sender', 'subject')

        Args:
            sender: Email address of the sender.
            to: Email address of the receiver.
            subject: The subject of the email message.
            message_text: The text of the email message.
            file_path: The path to the file to be attached.

        Returns:
            An object containing a base64url encoded email object.
        """
        Validator.is_file(file_path=file_path)

        message = MIMEMultipart()
        message['to'] = str(to)
        message['from'] = str(sender)
        message['subject'] = str(subject)
        if message_text is not None:
            msg = MIMEText(str(message_text))
            message.attach(msg)

        main_type, sub_type = cls._content_type(file_path)
        file_contents = cls._file_contents(file_path, main_type)
        attachment = cls._make_attachment(file_path, file_contents, main_type, sub_type)
        message.attach(attachment)

        return message

    @classmethod
    def create_text(cls, sender, to, subject, message_text):
        """Create a message for an email.

        Example:

        Example:

            >>> res = Mail.create_text('sender', 'a@a.com', 'subject', 'Message text')
            >>> res['to'], res['from'], res['subject']
            ('a@a.com', 'sender', 'subject')

        Args:
            sender: Email address of the sender.
            to: Email address of the receiver.
            subject: The subject of the email message.
            message_text: The text of the email message.

        Returns:
            An object containing a base64url encoded email object.
        """
        message = MIMEText(message_text)
        message['to'] = str(to)
        message['from'] = str(sender)
        message['subject'] = str(subject)

        return message



"""Home assistant related push plugins"""

from . import PushBase
from ...shared.hass import HassApi
from ...utils import auto_str_ignore


@auto_str_ignore(['_client', 'token'])
class Service(PushBase):
    """
    Calls a home assistant service providing the payload as service-data.

    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/push/hass.Service/index.md
    """
    __prefix__ = 'hass'

    def __init__(self, url, token, domain, service, timeout=10, **kwargs):
        super().__init__(**kwargs)
        self.url = str(url)
        self.token = str(token)
        self.domain = str(domain)
        self.service = str(service)
        self.timeout = timeout and int(timeout)
        if self.timeout <= 0:
            self.timeout = None  # Basically means no timeout
        self._client = HassApi(self.url, self.token, self.timeout)

    def _call(self, data):
        endpoint = 'services/{self.domain}/{self.service}'.format(**locals())

        try:
            return self._client.call(endpoint, method='post', data=data)
        except RuntimeError as exc:
            raise RuntimeError(
                "Failed to call the service {endpoint} @ {self.url}".format(**locals())
            ) from exc

    def push(self, payload):
        self._call(payload)
        return payload



from abc import abstractmethod
from datetime import datetime, timedelta

from .. import Plugin
from ...utils import parse_duration_literal


class UserDefinedFunction(Plugin):
    def __init__(self, throttle=None, **kwargs):
        super().__init__(**kwargs)
        self.throttle = throttle and parse_duration_literal(throttle)
        self._cache = None
        self._last_call = None

    def __call__(self, *args, **kwargs):
        if not self.throttle:
            return self.action(*args, **kwargs)

        # Invariant: self.throttle is not None
        n = datetime.now()
        if self._last_call is None:
            res = self.action(*args, **kwargs)
            self._cache = res
            self._last_call = n
            return res

        # Invariant: self._last_call is not None
        span = n - self._last_call
        if span >= timedelta(seconds=self.throttle):
            res = self.action(*args, **kwargs)
            self._cache = res
            self._last_call = n
            return res

        # Invariant: span < delta(throttle)
        return self._cache

    @abstractmethod
    def action(self, *args, **kwargs):
        raise NotImplementedError()



"""Useful sensor stuff."""

import logging
import os
from datetime import datetime, timedelta

import requests

from . import PullBase, Polling, PollingError
from .. import load_optional_module
from ...utils import safe_get, auto_str_ignore, parse_duration_literal, Throttle
from ...validator import Validator


class DHT(Polling):
    """
    Periodically polls a dht11 or dht22 (aka am2302) for temperature and humidity readings.
    Polling interval is controlled by `interval`.

    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/pull/sensor.DHT/index.md
    """
    __prefix__ = 'dht'

    EXTRA = 'dht'

    def __init__(self, device='dht22', data_gpio=17, humidity_offset=0.0, temp_offset=0.0,
                 **kwargs):
        super().__init__(**kwargs)
        valid_devices = ['dht11', 'dht22', 'am2302']
        self.device = str(device).lower()
        Validator.one_of(valid_devices, device=self.device)
        self.data_gpio = int(data_gpio)
        self.humidity_offset = float(humidity_offset)
        self.temp_offset = float(temp_offset)

    @staticmethod
    def _load_dht_package():
        try:
            import Adafruit_DHT as Pkg
        except ImportError:  # pragma: no cover
            logger = logging.getLogger(__name__)
            logger.warning("Adafruit_DHT package is not available - Using mock")
            from ...mocking import DHTMock as Pkg
        return Pkg

    def poll(self):
        # Adafruit package is optional - import at the last moment
        pkg = self._load_dht_package()

        device_map = {'dht11': pkg.DHT11, 'dht22': pkg.DHT22, 'am2302': pkg.AM2302}
        humidity, temperature = pkg.read_retry(device_map[self.device], self.data_gpio)

        if humidity is None or temperature is None:
            raise PollingError("Failed to get '{}' @ GPIO {} readings...".format(
                self.device, str(self.data_gpio)))

        return {
            'humidity': round(float(humidity) + self.humidity_offset, 2),
            'temperature': round(float(temperature) + self.temp_offset, 2)
        }


@auto_str_ignore(['_poller'])
class MiFlora(Polling):
    """
    Periodically polls a xiaomi miflora plant sensor for sensor readings via btle.

    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/pull/sensor.MiFlora/index.md

    """
    EXTRA = 'miflora'

    def __init__(self, mac, adapter='hci0', **kwargs):
        super().__init__(**kwargs)
        self.mac = str(mac)
        self.adapter = str(adapter)
        self._poller = None

    def _find_backend(self):
        try:
            import bluepy.btle  # noqa: F401 pylint: disable=unused-import
            from btlewrap import BluepyBackend
            backend = BluepyBackend
        except ImportError:
            from btlewrap import GatttoolBackend
            backend = GatttoolBackend
        self.logger.debug("Miflora for %s is using %s backend", self.mac, backend.__name__)
        return backend

    def _init(self):
        if self._poller:
            # We already got a miflora poller -> abort
            return

        mfp = load_optional_module('miflora.miflora_poller', self.EXTRA)
        self._poller = mfp.MiFloraPoller(
            mac=self.mac, adapter=self.adapter, backend=self._find_backend()
        )
        self.logger.debug("Initialization for %s finished", self.mac)

    def _connect_sensor(self):
        assert self._poller

        from btlewrap import BluetoothBackendException
        try:
            self.logger.debug("Reading miflora sensor: %s", self.mac)
            self._poller.fill_cache()
        except IOError as ioerr:
            raise PollingError() from ioerr
        except BluetoothBackendException as bterror:
            raise PollingError() from bterror

    def _get_sensor_readings(self):
        assert self._poller

        mfp = load_optional_module('miflora.miflora_poller', self.EXTRA)
        reading_params = [mfp.MI_CONDUCTIVITY, mfp.MI_LIGHT, mfp.MI_MOISTURE,
                          mfp.MI_TEMPERATURE, mfp.MI_BATTERY]
        res = {para: self._poller.parameter_value(para) for para in reading_params}
        return {**res, **{'firmware': self._poller.firmware_version()}}

    def poll(self):
        self._init()
        self._connect_sensor()
        readings = self._get_sensor_readings()
        self.logger.info("Miflora readings for '%s': %s", self.mac, readings)
        return readings


@auto_str_ignore(ignore_list=["api_key"])
class OpenWeather(Polling):
    """
    Periodically polls weather data from the OpenWeatherMap api.

    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/pull/sensor.OpenWeather/index.md
    """
    __prefix__ = 'openweather'

    def __init__(self, api_key, lat=None, lon=None, city_name=None, units="metric", tz=None,
                 **kwargs):  # pylint: disable=too-many-arguments
        super().__init__(**kwargs)
        self.api_key = str(api_key)
        self.lat = Validator.cast_or_none(float, lat)
        self.lon = Validator.cast_or_none(float, lon)
        self.city_name = Validator.cast_or_none(str, city_name)
        if self.city_name is None and not self._validate_lat_lon():
            raise ValueError("You have to pass city_name or lat and lon.")

        Validator.one_of(["metric", "imperial", "kelvin"], units=units)
        self.units = units
        self.tzone = Validator.cast_or_none(str, tz)

        from pytz import timezone
        from tzlocal import get_localzone
        self._tz = get_localzone() if self.tzone is None else timezone(self.tzone)

    def poll(self):
        url = self._create_request_url()
        resp = requests.get(url)
        if resp.status_code != 200:
            raise PollingError("GET of '{url}' failed with status "
                               "code = '{resp.status_code}'".format(**locals()))

        raw_data = resp.json()

        return dict(
            temperature=safe_get(raw_data, "main", "temp"),
            pressure=safe_get(raw_data, "main", "pressure"),
            humidity=safe_get(raw_data, "main", "humidity"),
            cloudiness=safe_get(raw_data, "clouds", "all"),
            wind=safe_get(raw_data, "wind"),
            poll_dts=datetime.now(self._tz).isoformat(),
            last_updated_dts=datetime.fromtimestamp(
                safe_get(raw_data, "dt"), tz=self._tz).isoformat(),
            raw=raw_data
        )

    def _validate_lat_lon(self):
        return self.lat is not None and self.lon is not None

    def _create_request_url(self):
        url = ("http://api.openweathermap.org/data/2.5/weather?"
               "units={self.units}&"
               "APPID={self.api_key}")
        if self.lat is None and self.lon is None:
            url += "&q={self.city_name}"
        else:
            url += "&lat={self.lat}&lon={self.lon}"
        return url.format(**locals())


# pylint: disable=too-many-instance-attributes
@auto_str_ignore(ignore_list=['cool_down_secs', 'notify', 'signal_fft', 'signal_length'])
class Sound(PullBase):
    """
    Listens to the microphone in realtime and searches the stream for a specific sound pattern.
    Practical example: I use this plugin to recognize my doorbell without tampering with
    the electrical device ;-)

    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/pull/sensor.Sound/index.md
    """
    EXTRA = 'sound'

    MODE_PEARSON = 'pearson'
    MODE_STD = 'std'
    ALLOWED_MODES = [MODE_PEARSON, MODE_STD]
    RATE = 44100
    CHUNK_SIZE = 1024 * 4
    PEARSON_THRESHOLD = 0.5
    STD_THRESHOLD = 1.4

    # pylint: disable=too-many-arguments
    def __init__(self, wav_file, device_index=None, mode='pearson', sensitivity_offset=0.0,
                 cool_down="10s", ignore_overflow=False, **kwargs):  # pylint: disable=too-many-arguments
        super().__init__(**kwargs)
        self.wav_file = str(wav_file)
        if not os.path.isabs(self.wav_file):
            self.wav_file = os.path.join(self.base_path, self.wav_file)
        Validator.is_file(wav_file=self.wav_file)
        self.device_index = device_index and int(device_index)
        self.wav_file_name = os.path.basename(os.path.splitext(self.wav_file)[0])
        self.signal_length, self.signal_fft = self._load_wav_fft()
        self.mode = str(mode)
        Validator.one_of(self.ALLOWED_MODES, mode=self.mode)
        self.sensitivity_offset = float(sensitivity_offset)
        self.cool_down = cool_down
        self.cool_down_secs = cool_down and parse_duration_literal(cool_down)  # Might be None
        self.ignore_overflow = bool(ignore_overflow)
        # Override notify with a throttled version
        self.notify = Throttle(timedelta(seconds=self.cool_down_secs or 0))(self.notify)

    def _load_wav_fft(self):
        wavfile = load_optional_module('scipy.io.wavfile', self.EXTRA)
        self.logger.debug("[%s] Loading wav file from '%s'", self.name, self.wav_file)
        sample_rate, signal = wavfile.read(self.wav_file)
        N, secs, signal_fft = self._perform_fft(signal, sample_rate)
        self.logger.debug("[%s] Loaded %s seconds wav file @ %s hz", self.name, secs, sample_rate)
        return N, signal_fft

    def _perform_fft(self, signal, rate, add_zeros=True):
        np = load_optional_module('numpy', self.EXTRA)
        scipy = load_optional_module('scipy', self.EXTRA)

        chn = len(signal.shape)
        if chn >= 2:  # Make mono channel
            signal = signal.sum(axis=1) / 2
        N = int(signal.shape[0])
        secs = N / float(rate)
        if add_zeros:
            signal = np.concatenate((signal, np.zeros(len(signal))), axis=None)
        trans_fft = abs(scipy.fft(signal))

        return N, secs, trans_fft

    def _similarity(self, buffer):
        _, _, current_fft = self._perform_fft(buffer[:self.signal_length], self.RATE, True)
        if self.mode == self.MODE_PEARSON:
            pearsonr = load_optional_module('scipy.stats.stats', self.EXTRA).pearsonr
            corrcoef = pearsonr(self.signal_fft, current_fft)[0]
            threshold = self.PEARSON_THRESHOLD + self.sensitivity_offset
        else:
            np = load_optional_module('numpy', self.EXTRA)
            corrcoef = np.correlate(
                self.signal_fft / self.signal_fft.std(),
                current_fft / current_fft.std()
            )[0] / len(self.signal_fft)
            threshold = self.STD_THRESHOLD + self.sensitivity_offset

        return corrcoef >= threshold, corrcoef, threshold

    def pull(self):
        np = load_optional_module('numpy', self.EXTRA)
        pyaudio = load_optional_module('pyaudio', self.EXTRA)

        pa = pyaudio.PyAudio()
        stream = pa.open(
            format=pyaudio.paInt16,
            input_device_index=self.device_index,
            channels=1,
            rate=self.RATE,
            input=True,
            frames_per_buffer=self.CHUNK_SIZE
        )
        try:
            buffer = None
            N = self.signal_length
            while not self.stopped:
                data = np.fromstring(stream.read(
                    self.CHUNK_SIZE,
                    exception_on_overflow=self.ignore_overflow
                ), dtype=np.int16)
                buffer = data if buffer is None else np.concatenate((buffer, data), axis=None)
                lbuf = len(buffer)
                if lbuf >= N:
                    self.logger.debug("[%s] Buffer (%s) >= size of wav file (%s)",
                                      self.name, lbuf, N)
                    flag, corrcoef, threshold = self._similarity(buffer)
                    self.logger.debug("[%s] Correlation: %s >= %s = %s",
                                      self.name, corrcoef, threshold, flag)
                    if flag:
                        self.notify({
                            'data': self.wav_file_name,
                            'corrcoef': corrcoef,
                            'threshold': threshold
                        })
                        buffer = None
                    else:
                        buffer = buffer[int(len(buffer) * 0.75):]
        finally:
            stream.close()
            pa.terminate()



"""Pull 'n' Push GMail Token Creator

Usage:
  pnp_gmail <credential_file> <token_file>
  pnp_gmail (-h | --help)

Options:
  -h --help         Show this screen.

Args:
  credential_file   Obtained from GMail
  token_file        Output tokens (existing file will be overwritten)

"""
import pickle

import os

import sys
from docopt import docopt
from google_auth_oauthlib.flow import InstalledAppFlow

# If modifying these scopes, delete the token file
SCOPES = ['https://www.googleapis.com/auth/gmail.send']


def run(credential_file, token_file):
    if not os.path.isfile(credential_file):
        print("Credential file '{}' does not exist".format(credential_file))
        return 1

    flow = InstalledAppFlow.from_client_secrets_file(credential_file, SCOPES)
    creds = flow.run_local_server()

    with open(token_file, 'wb') as tfp:
        pickle.dump(creds, tfp)

    print("Tokens successfully written to: {}".format(token_file))
    return 0


def main():
    arguments = docopt(__doc__)
    creds, token = arguments['<credential_file>'], arguments['<token_file>']
    return run(creds, token)


if __name__ == '__main__':
    sys.exit(main())



import warnings
from functools import partial

from . import PushBase
from ...shared.exc import TemplateError
from ...utils import parse_duration_literal, make_list
from ...validator import Validator


class Echo(PushBase):
    """
    This push simply logs the `payload` via the `logging` module.

    Examples:

        >>> dut = Echo(name="echo_push")
        >>> dut.push("I will be logged")
        'I will be logged'
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def push(self, payload):
        envelope, real_payload = self.envelope_payload(payload)
        self.logger.info("[{self.name}] Got '{real_payload}' with envelope '{envelope}'".format(**locals()))
        return payload  # Payload as is. With envelope (if any)


class Nop(PushBase):
    """
    Executes no operation at all. A call to push(...) just returns the payload.
    This push is useful when you only need the power of the selector for dependent pushes.

    Nop = No operation OR No push ;-)

    Examples:

        >>> dut = Nop(name="nop_push")
        >>> dut.push('I will be returned unaltered')
        'I will be returned unaltered'
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.last_payload = None

    def push(self, payload):
        self.last_payload = payload
        return payload


class TemplatedExecute(PushBase):
    """
    Executes a command with given arguments in a shell of the operating system.
    Both `command` and `args` may include placeholders (e.g. `{{placeholder}}`) which are injected at runtime
    by passing the specified payload after selector transformation. Please see the Examples section for further details.

    Will return the exit code of the command and optionally the output from stdout and stderr.

    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/push/simple.TemplatedExecute/index.md
    """
    def __init__(self, command, args=None, cwd=None, capture=True, timeout="5s", **kwargs):
        super().__init__(**kwargs)
        self._command = str(command) if command else None
        self._args = self._parse_args(args)
        self._cwd = str(cwd) if cwd else self.base_path
        Validator.is_directory(allow_none=True, cwd=self._cwd)
        self._capture = bool(capture)
        self._timeout = timeout and parse_duration_literal(timeout)

    @staticmethod
    def _render_jinja_template(template, subs):
        from jinja2 import StrictUndefined
        from jinja2 import Template
        from jinja2.exceptions import UndefinedError
        try:
            tpl = Template(template, undefined=StrictUndefined)
            return tpl.render(**(subs or {}))
        except UndefinedError as exc:
            raise TemplateError('Error when rendering template in TemplatedExecute') from exc

    def _parse_args(self, val):
        args = make_list(val)
        if args:
            return [str(arg) for arg in args]

    def _serialize_args(self, transform_fun=None, add_quotes=True):
        def escape_fun(arg):
            return str(arg).replace('"', '\\"')

        def quotes_fun(arg):
            if arg is None or len(str(arg)) == 0 or str(arg).strip() == 'None':
                return ''
            return '"{}"'.format(escape_fun(arg)) if add_quotes else str(arg)

        args = self._args
        if not args:
            return None
        if transform_fun:
            args = [transform_fun(arg) for arg in args]
        # Do argument quoting
        args = [quotes_fun(arg) for arg in args]
        # Remove empty args
        args = [arg for arg in args if arg != '']
        return " ".join(args)

    def _execute(self, command_str):
        def _output(br):
            return [line.strip('\n\r') for line in br]

        import subprocess
        self.logger.info("[{self.name}] Running command in shell: {command_str}".format(**locals()))
        p = subprocess.Popen(
            args=command_str,
            shell=True,
            cwd=self._cwd,
            universal_newlines=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )
        try:
            if self._timeout:
                p.wait(timeout=int(self._timeout))

            res = dict(return_code=p.returncode)
            if self._capture:
                res['stdout'] = _output(p.stdout)
                res['stderr'] = _output(p.stderr)
            return res
        finally:
            p.stdout.close()
            p.stderr.close()

    def push(self, payload):
        if isinstance(payload, dict):
            subs = payload
        else:
            subs = dict(data=payload, payload=payload)

        command_str = self._render_jinja_template(self._command, subs=subs)
        if self._args:
            args = self._serialize_args(
                transform_fun=partial(self._render_jinja_template, subs=subs),
                add_quotes=True
            )
            command_str = "{command_str} {args}".format(**locals())

        return self._execute(command_str)


class Execute(TemplatedExecute):
    """
    Executes a command with given arguments in a shell of the operating system.

    Will return the exit code of the command and optionally the output from stdout and stderr.

    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/push/simple.Execute/index.md
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        warnings.simplefilter('always', DeprecationWarning)
        warnings.warn(
            "Plugin `pnp.plugins.push.simple.Execute` is deprecated."
            " Please use `pnp.plugins.push.simple.TemplatedExecute`."
            " In the near future `Execute` will be replaced by `TemplatedExecute`.",
            category=DeprecationWarning,
            stacklevel=2
        )
        warnings.simplefilter('default', DeprecationWarning)

    def push(self, payload):
        envelope, real_payload = self.envelope_payload(payload)
        args = self._parse_envelope_value('args', envelope)  # Override args via envelope

        command_str = self._command
        if args:
            args = self._serialize_args(
                transform_fun=None,
                add_quotes=False
            )
            command_str = "{command_str} {args}".format(**locals())

        return self._execute(command_str)



"""Home assistant related utility classes"""

import json

import attr


@attr.s
class HassApi:
    """Utility class to communicate with home assistant via the rest-api."""

    METHOD_GET = 'get'
    METHOD_POST = 'post'
    ALLOWED_METHODS = [METHOD_GET, METHOD_POST]

    base_url = attr.ib(type=str, converter=str)
    token = attr.ib(type=str, converter=str)
    timeout = attr.ib(type=int, converter=lambda x: int(x) if x is not None else None, default=None)

    def call(self, endpoint, method='get', data=None):
        """Calls the specified endpoint (without prefix api) using the given method.
        You can optionally pass data to the request which will be json encoded."""
        from requests import get, post
        import urllib.parse as urlparse

        method = str(method).lower()
        if method not in self.ALLOWED_METHODS:
            raise ValueError(
                "Argument method is expected to be one of {allowed}, but is '{method}'".format(
                    allowed=str(self.ALLOWED_METHODS),
                    method=method)
            )

        url = urlparse.urljoin(urlparse.urljoin(self.base_url, 'api/'), endpoint)
        headers = {
            'Authorization': 'Bearer {self.token}'.format(**locals()),
            'content-type': 'application/json',
        }

        if data is not None:
            data = json.dumps(data)

        if method == self.METHOD_GET:
            response = get(url, headers=headers, timeout=self.timeout, data=data)
        else:
            response = post(url, headers=headers, timeout=self.timeout, data=data)

        if response.status_code != 200:
            raise RuntimeError("Failed to call endpoint {url}"
                               "\nHttp Code: {response.status_code}"
                               "\nMessage: {response.text}".format(**locals()))

        return response.json()
